# -*- coding: utf-8 -*-
"""
Copyright (C) 2015-2016 Benjamin Eltzner

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 3
of the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
or see <http://www.gnu.org/licenses/>.
"""

import math, warnings, sys, os
import numpy as np
import numpy.random as arandom
import numpy.linalg as la
import matplotlib.pyplot as plot
import scipy.optimize as opt
#from sympy import symbols, Eq, solve, exp
from io_pkg import import_csv, import_lists, find_files, export_csv
from mpl_toolkits.mplot3d import proj3d
import scipy.stats as stat

"""#############################################################################
################################   Constants   #################################
#############################################################################"""

PI = np.pi
DEG = math.degrees(1)
RAD = math.radians(1)
EPS = 1e-10
AX = np.newaxis
AXES_LABEL_SIZE = 16.0
GELS = ['01kPa','10kPa','30kPa']
OUT = 'out/'
def orthogonal_proj(zfront, zback):
    a = (zfront+zback)/(zfront-zback)
    b = -2*(zfront*zback)/(zfront-zback)
    return np.array([[1,0,0,0], [0,1,0,0], [0,0,a,b], [0,0,-1e-4,zback]])
proj3d.persp_transformation = orthogonal_proj

"""#############################################################################
############################   Numerics functions   ############################
#############################################################################"""

"""
This function feeds a function into scipy's standard optimization algorithms.
Due to hidden increases of rounding error, the fits may not converge
sufficiently precisely. In these cases, we increase error tolerance (starting at
1:10^(-8) and at most to increasing to 1:0.01) until the fit converges.
"""
def fit (f, d, initial, factor, shift, verbose):
    warnings.filterwarnings('error')
    initial = factor * initial - shift
    tol = 1e-8
    if verbose:
        print('Starting fit... ',)
        sys.stdout.flush()
    try:
        initial, exit_code = opt.leastsq(f, initial)
    except:
        if verbose: print('Exception in fit! Setting invalid exit code.')
        exit_code = 6
    fails = 0
    counter = 0
    while exit_code > 1:
        if verbose:
            print('failed fit with exit code:', exit_code, '\nRestarting... ')
        fails += 1
        if fails > 3 or exit_code == 6:
            if verbose:
                print('with new start value... ')
            initial = factor * arandom.rand(d) - shift
            fails = 0
        try:
            initial, exit_code = opt.leastsq(f, initial, ftol=tol, xtol=tol)
        except:
            if verbose: print('Exception in fit! Setting invalid exit code.')
            exit_code = 6
        if counter > 20:
            return None
        counter += 1
        tol *= 2
    if verbose:
        print('done! Exit code:', exit_code, '\n', initial)
        sys.stdout.flush()
    return initial

def piecewise_linear_fit (x):
    """
    Function f: R^(d+1) -> R^N for which the Levenberg-Marquardt method is
    applied to identify the argument which minimizes the sum of squares.
    This function models the piecewise linear curve.
    """
    def f (v):
        if v is None:
            #print('Oh noes!', flush=True)
            return 1.001 * x
        if (v[1] < 1) or (v[1] < v[0]) or (v[1] > len(x)) or (v[0] < -15):
            return (2 + max(1-v[1], (v[0] - v[1]), (v[1] - len(x)), (15 - v[0]))) * x
        t0 = max(int(v[0]), 0)
        x1 = x[:t0].copy()
        x2 = x[t0:int(v[1])].copy()
        x3 = x[int(v[1]):].copy()
        x2 -= (np.array(range(len(x2))) - v[0] + t0 + 1) * v[2]
        x3 -= ((v[1] - v[0]) * v[2] +
               (np.array(range(len(x3))) - v[1] + int(v[1]) + 1) * v[3])
        out = np.hstack((x1,x2,x3))
        return out
    results = []
    values = []
    """ We fit 10 times and take the best fit to avoid local minima. """
    for i in range(10):
        fv = 2 * x
        while (la.norm(fv) > 0.999 * la.norm(x)):
            v = fit(f, 4, arandom.rand(4), np.array([1,100,10,1]), 0, False)
            fv = f(v)
        results.append((v, fv))
        values.append(la.norm(fv))
    return results[np.argmin(values)]

"""
In this function, we impose an upper bound on the variance of the fit. This
can be used to discard fits that run into a local minimum.
"""
def piecewise_linear_fit_with_bound (x, sigma, par):
    """
    Function f: R^(d+1) -> R^N for which the Levenberg-Marquardt method is
    applied to identify the argument which minimizes the sum of squares.
    """
    def f (v):
        if v is None:
            #print('Oh noes!', flush=True)
            return x
        if (v[1] < 1) or (v[1] < v[0]) or (v[1] > len(x)) or (v[0] < -15):
            return (2 + max(1-v[1], (v[0] - v[1]), (v[1] - len(x)), (15 - v[0]))) * x
        t0 = max(int(v[0]), 0)
        x1 = x[:t0].copy()
        x2 = x[t0:int(v[1])].copy()
        x3 = x[int(v[1]):].copy()
        x2 -= (np.array(range(len(x2))) - v[0] + t0 + 1) * v[2]
        x3 -= ((v[1] - v[0]) * v[2] +
               (np.array(range(len(x3))) - v[1] + int(v[1]) + 1) * v[3])
        out = np.hstack((x1,x2,x3))
        return out
    fv = 2 * x
    while (np.std(fv) > sigma):
        v = fit(f, 4, arandom.rand(4), 2 * par, 0, False)
        fv = f(v)
    return v, fv

"""
This funtion is used for the non-parametric bootstrap.
"""
def piecewise_linear_fit_disordered (x, y, sigma, par):
    """
    Function f: R^(d+1) -> R^N for which the Levenberg-Marquardt method is
    applied to identify the argument which minimizes the sum of squares.
    """
    def f (v):
        if v is None:
            #print('Oh noes!', flush=True)
            return y
        if (v[1] < 1) or (v[1] < v[0]) or (v[1] > len(x)) or (v[0] < -15):
            return (2 + max(1-v[1], (v[0] - v[1]), (v[1] - len(x)), (15 - v[0]))) * y
        t0 = max(int(v[0]), 0)
        x1 = np.zeros(t0)
        x2 = (np.array(range(int(v[1]) - t0)) - v[0] + t0 + 1) * v[2]
        x3 = ((v[1] - v[0]) * v[2] +
              (np.array(range(len(x) - int(v[1])) - v[1] + int(v[1]) + 1) * v[3]))
        fit_line = np.hstack((x1,x2,x3))
        out = y - fit_line[x]
        return out
    fv = 2 * y
    while (np.std(fv) > sigma):
        v = fit(f, 4, arandom.rand(4), 2 * par, 0, False)
        fv = f(v)
    return v, fv

"""
This function performs the fit for a fixed t1=n.
"""
def piecewise_linear_fit2 (x, n):
    #print(n, flush=True)
    """
    Function f: R^(d+1) -> R^N for which the Levenberg-Marquardt method is
    applied to identify the argument which minimizes the sum of squares.
    """
    def f (v):
        if v is None:
            #print('Oh noes!', flush=True)
            return 1.001 * x 
        if (v[0] > n) or ((n - v[0]) * v[1] < 5) or (v[0] < -10000):
            return (2 + max((v[0] - n), (5 - (n - v[0]) * v[1]), (-10000 - v[0]))) * x
        t0 = max(int(v[0]), 0)
        x1 = x[:t0].copy()
        x2 = x[t0:n].copy()
        x3 = x[n:].copy()
        x2 -= (np.array(range(len(x2))) - v[0] + t0 + 1) * v[1]
        x3 -= ((n - v[0]) * v[1] + (np.array(range(len(x3))) + 1) * v[2])
        out = np.hstack((x1,x2,x3))
        return out
    results = []
    values = []
    for i in range(10):
        fv = 2 * x
        while (la.norm(fv) > la.norm(x)):
            v = fit(f, 3, arandom.rand(3), np.array([-15,10/n,0.1]), 0, False)
            fv = f(v)
        results.append((v, fv))
        values.append(la.norm(fv))
    return results[np.argmin(values)]

"""
This function performs the fit for a fixed t1=n and t0=-150s.
"""
def piecewise_linear_fit3 (x, n):
    #print(n, flush=True)
    """
    Function f: R^(d+1) -> R^N for which the Levenberg-Marquardt method is
    applied to identify the argument which minimizes the sum of squares.
    """
    def f (v):
        if v is None:
            #print('Oh noes!', flush=True)
            return 1.001 * x
        x2 = x[:n].copy()
        x3 = x[n:].copy()
        x2 -= (np.array(range(len(x2))) + 15) * v[0]
        x3 -= ((n + 15) * v[0] + (np.array(range(len(x3))) + 1) * v[1])
        out = np.hstack((x2,x3))
        return out
    results = []
    values = []
    for i in range(10):
        fv = 2 * x
        while (la.norm(fv) > la.norm(x)):
            v = fit(f, 2, arandom.rand(2), np.array([10/n,0.1]), 0, False)
            fv = f(v)
        results.append((np.hstack((-15, v)), fv))
        values.append(la.norm(fv))
    return results[np.argmin(values)]

def parametric_bootstrap (par, n, sigma, s):
    t0 = max(int(par[0]), 0)
    x1 = np.zeros(t0)
    x2 = (np.array(range(int(par[1]) - t0)) - par[0] + t0 + 1) * par[2]
    x3 = ((par[1] - par[0]) * par[2] +
          (np.array(range(n - int(par[1])) - par[1] + int(par[1]) + 1) * par[3]))
    lines = np.tile(np.hstack((x1,x2,x3)), s).reshape((s,n))
    noise = sigma * arandom.randn(n*s).reshape((s,n))
    return lines + noise, np.std(noise, axis=1)

def nonparametric_bootstrap (data, noise, s):
    n = len(data)
    numbers = arandom.randint(n, size=n*s)
    lines = data[numbers].copy().reshape((s,n))
    noise = noise[numbers].copy().reshape((s,n))
    return numbers.reshape((s,n)), lines, np.sqrt(np.mean(noise**2, axis=1))

"""
Plot functions
"""

def plot_line (data, v, filename=None):
    fig = plot.figure()
    diag = fig.add_subplot(111)
    #diag.set_xlim(left=0)
    diag.scatter(10*np.array(range(len(data))), data, lw=0, s=5, color='#aaaaaa')
    t0 = max(int(v[0]), 0)
    x1 = np.zeros(t0)
    x2 = (np.array(range(int(v[1]) - t0)) - v[0] + t0 + 1) * v[2]
    x3 = ((v[1] - v[0]) * v[2] +
               (np.array(range(len(data) - int(v[1])) - v[1] + int(v[1]) + 1) * v[3]))
    fit_line = np.hstack((x1,x2,x3))
    #print(len(data), len(fit_line))
    diag.plot(10*np.array(range(len(data))), fit_line, lw=2, color='black')
    diag.set_xlim(left=0)
    diag.set_ylim(bottom=0)
    if filename != None:
        plot.savefig(OUT + filename + '.png', dpi= 300)
    else:
        plot.show()
    plot.close()

def plot_series (x, y, filename=None):
    fig = plot.figure()
    diag = fig.add_subplot(111)
    #diag.set_xlim(left=0)
    diag.plot(x, y)
    diag.set_xlim(left=0)
    #diag.set_ylim(top=min(max(y), 0.35))
    if filename != None:
        plot.savefig(OUT + filename + '.png', dpi= 300)
    else:
        plot.show()
    plot.close()

def plot_histogram (data, filename=None):
    k = 1
    x = np.arange(0, 600, k)
    y = np.histogram(data, bins=int(600/k), range=(0,600))[0]
    fig = plot.figure()
    diag = fig.add_subplot(111)
    diag.bar(10*x, y, width=10*k, linewidth=0)
    diag.set_xlim((min(data) // 10) * 100, (max(data) // 10 + 1) * 100)
    if filename != None:
        plot.savefig(OUT + filename + '.png', dpi= 300)
    else:
        plot.show()
    plot.close()

"""#############################################################################
#############################   Control functions   ############################
#############################################################################"""

"""
This function is prototypical for all functions in this section.
"""
def fit_some_data (folder):
    files = find_files('*.csv', 'data/total_length_time_series/' + folder)
    times = list(range(10,600,5))
    names = []
    out = []

    """ Loop over the time series. """
    for f in sorted(files):
        name = os.path.basename(f)[:-4].replace(' ', '_')
#        if name != '010216-P6':
#            continue
        names.append(name)
        print(name, flush=True)
        data = import_csv(f)[:,1] * 0.109 # Lengths in micrometers
        values = []

        """ Plot goodness of fit for t1 ranging from 100s to 6000s. """
        for n in times:
            if n > len(data) - 10:
                continue
            v, fv = piecewise_linear_fit2(data, n)
            if v[0] < -15:
                v, fv = piecewise_linear_fit3(data, n)
            values.append(np.std(fv))
            print(n, end=' ', flush=True)
        print()
        plot_series([10*t for t in times[:len(values)]], values,
                    name + '_t1_standard_deviations')

        """ Actually perform the fit and plot it. """
        v, fv = piecewise_linear_fit(data)
        plot_line(data, v, name + '_best_fit')

        """ Parametric bootstrap """
        boots = []
        lines, sigmas = parametric_bootstrap(v, len(data), np.std(fv), 1000)
        for i,l in enumerate(lines):
            if (i+1) % 100 == 0:
                print(i+1, end=' ', flush=True)
            vl, fvl = piecewise_linear_fit_with_bound(l, 1.01 * sigmas[i], v)
            boots.append(vl[1])
        print()
        plot_histogram(boots, name + '_parametric_bootstrap')
        v = np.hstack((v, np.std(fv), 10*np.mean(boots), 10*np.std(boots)))

        """ Non-parametric bootstrap """
        boots = []
        numbers, lines, sigmas = nonparametric_bootstrap(data, fv, 1000)
        for i,l in enumerate(lines):
            if (i+1) % 100 == 0:
                print(i+1, end=' ', flush=True)
            vl, fvl = piecewise_linear_fit_disordered(numbers[i], l,
                                                      1.01 * sigmas[i], v[:4])
            boots.append(vl[1])
        print()
        plot_histogram(boots, name + '_nonparametric_bootstrap')
        v = np.hstack((v, 10*np.mean(boots), 10*np.std(boots)))

        """ One time step equals 10s. """
        v[:2] *= 10
        v[2:4] *= 0.1
        out.append(v)
        print(v, flush=True)

    """ Save fit results. """
    output = open('out/parameters_' + folder + '.csv', 'w')
    output.write('%s\n' % ','.join(names))
    output.write('\n'.join(','.join("%.6f" % i for i in k)
                           for k in np.array(out).T) + '\n')
    output.close()

def fit_data ():
    files = find_files('*.csv', 'data/total_length_time_series')
    for f in sorted(files):
        data = import_csv(f)[:,1]
        v, fv = piecewise_linear_fit(data)
        plot_line(data,v)
        print(f, v, la.norm(fv), la.norm(data))

def fit_data2 ():
    files = find_files('*.csv', 'data/total_length_time_series/old')
    print(files, len(files))
    times = list(range(10,600,5))
    names = []
    out = []
    for f in sorted(files):
        name = os.path.basename(f)[:-4].replace(' ', '_')
#        if name != '120716_onlyactin_P3':
#            continue
        names.append(name)
        print(name, flush=True)
        data = import_csv(f)[:,1] * 0.109 # Lengths in micrometers
        if name == '120716_onlyactin_P3':
            data = data[121:]
        values = []
        for n in times:
            if n > len(data) - 10:
                continue
            v, fv = piecewise_linear_fit2(data, n)
            if v[0] < -15:
                v, fv = piecewise_linear_fit3(data, n)
            values.append(np.std(fv))
            print(n, end=' ', flush=True)
        print()
        plot_series([10*t for t in times[:len(values)]], values,
                    name + '_t1_standard_deviations')
        v, fv = piecewise_linear_fit(data)
        plot_line(data, v, name + '_best_fit')
        boots = []
        lines, sigmas = parametric_bootstrap(v, len(data), np.std(fv), 1000)
        for i,l in enumerate(lines):
            if (i+1) % 100 == 0:
                print(i+1, end=' ', flush=True)
            vl, fvl = piecewise_linear_fit_with_bound(l, 1.01 * sigmas[i], v)
            boots.append(vl[1])
        print()
        plot_histogram(boots, name + '_parametric_bootstrap')
        v = np.hstack((v, np.std(fv), 10*np.mean(boots), 10*np.std(boots)))
        boots = []
        numbers, lines, sigmas = nonparametric_bootstrap(data, fv, 1000)
        for i,l in enumerate(lines):
            if (i+1) % 100 == 0:
                print(i+1, end=' ', flush=True)
            vl, fvl = piecewise_linear_fit_disordered(numbers[i], l,
                                                      1.01 * sigmas[i], v[:4])
            boots.append(vl[1])
        print()
        plot_histogram(boots, name + '_nonparametric_bootstrap')
        v = np.hstack((v, 10*np.mean(boots), 10*np.std(boots)))
        v[:2] *= 10
        v[2:4] *= 0.1
        out.append(v)
        print(v, flush=True)
    output = open('out/parameters.csv', 'w')
    output.write('%s\n' % ','.join(names))
    output.write('\n'.join(','.join("%.6f" % i for i in k)
                           for k in np.array(out).T) + '\n')
    output.close()

def fit_area ():
    files = find_files('*.csv', 'data/area_time_series')
    times = list(range(1,600,5))
    out = {}
    for f in sorted(files):
        name = 'area_' + os.path.basename(f)[:-4]
        data = import_csv(f)
        print(name, data.shape, flush=True)
        values = []
        for n in times:
            if n > len(data) - 10:
                continue
            v, fv = piecewise_linear_fit2(data, n)
            values.append(np.std(fv))
        plot_series([10*t for t in times[:len(values)]], values,
                    name + '_t1_standard_deviations')
        v, fv = piecewise_linear_fit(data)
        plot_line(data, v, name + '_best_fit')
        boots = []
        lines, sigmas = parametric_bootstrap(v, len(data), np.std(fv), 1000)
        for i,l in enumerate(lines):
            if (i+1) % 100 == 0:
                print(i+1, end=' ', flush=True)
            vl, fvl = piecewise_linear_fit_with_bound(l, 1.01 * sigmas[i], v)
            boots.append(vl[1])
        print()
        plot_histogram(boots, name + '_parametric_bootstrap')
        v = np.hstack((v, np.std(fv), 10*np.mean(boots), 10*np.std(boots)))
        boots = []
        numbers, lines, sigmas = nonparametric_bootstrap(data, fv, 1000)
        for i,l in enumerate(lines):
            if (i+1) % 100 == 0:
                print(i+1, end=' ', flush=True)
            vl, fvl = piecewise_linear_fit_disordered(numbers[i], l,
                                                      1.01 * sigmas[i], v[:4])
            boots.append(vl[1])
        print()
        plot_histogram(boots, name + '_nonparametric_bootstrap')
        v = np.hstack((v, 10*np.mean(boots), 10*np.std(boots)))
        v[:2] *= 10
        v[2:4] *= 0.1
        print(v[:4], np.std(fv), v[4:], flush=True)
        out[name] = v
    export_csv(out, 'area_parameters.csv')

def fit_data_sigmas ():
    files = find_files('*.csv', 'data/total_length_time_series')
    for f in sorted(files):
        name = os.path.basename(f)[:-4]
        data = import_csv(f)[:,1] * 0.109 # Lengths in micrometers
        v, fv = piecewise_linear_fit(data)
        print(name, np.std(fv), flush=True)

def fit_new_data ():
    files = find_files('*.csv', 'data/total_length_time_series/new')
    times = list(range(10,600,5))
    names = []
    out = []
    for f in sorted(files):
        name = os.path.basename(f)[:-4]
#        if name != '010216-P6':
#            continue
        names.append(name)
        print(name, flush=True)
        data = import_csv(f)[:,1] * 0.109 # Lengths in micrometers
        values = []
        for n in times:
            if n > len(data) - 10:
                continue
            v, fv = piecewise_linear_fit2(data, n)
            if v[0] < -15:
                v, fv = piecewise_linear_fit3(data, n)
            values.append(np.std(fv))
            print(n, end=' ', flush=True)
        print()
        plot_series([10*t for t in times[:len(values)]], values,
                    name + '_t1_standard_deviations')
        v, fv = piecewise_linear_fit(data)
        plot_line(data, v, name + '_best_fit')
        boots = []
        lines, sigmas = parametric_bootstrap(v, len(data), np.std(fv), 1000)
        for i,l in enumerate(lines):
            if (i+1) % 100 == 0:
                print(i+1, end=' ', flush=True)
            vl, fvl = piecewise_linear_fit_with_bound(l, 1.01 * sigmas[i], v)
            boots.append(vl[1])
        print()
        plot_histogram(boots, name + '_parametric_bootstrap')
        v = np.hstack((v, np.std(fv), 10*np.mean(boots), 10*np.std(boots)))
        boots = []
        numbers, lines, sigmas = nonparametric_bootstrap(data, fv, 1000)
        for i,l in enumerate(lines):
            if (i+1) % 100 == 0:
                print(i+1, end=' ', flush=True)
            vl, fvl = piecewise_linear_fit_disordered(numbers[i], l,
                                                      1.01 * sigmas[i], v[:4])
            boots.append(vl[1])
        print()
        plot_histogram(boots, name + '_nonparametric_bootstrap')
        v = np.hstack((v, 10*np.mean(boots), 10*np.std(boots)))
        v[:2] *= 10
        v[2:4] *= 0.1
        out.append(v)
        print(v, flush=True)
    output = open('out/parameters_new.csv', 'w')
    output.write('%s\n' % ','.join(names))
    output.write('\n'.join(','.join("%.6f" % i for i in k)
                           for k in np.array(out).T) + '\n')
    output.close()

if __name__ == "__main__":
    #fit_data2()
    #fit_area()
    #fit_data_sigmas()
    #fit_new_data()
    fit_some_data('problem')